// ********************************************************************
// Demoprogramm
// Inhalt: LCD (DOG-M 162 2 Zeilen) mit (ST7036 Controller) ansteuern,
// im 8-Bit parallel-Mode
// LCD-Controller: ST7036 kompatibel
//                 fr HD44780 sind nderungen der Initialisierung
//                 und der Zeilenadressenzuordnung notwendig
// Stand: 17.10.2007
// SDCC-Version: M-IDE-51/SDCC 2.7.0
// AT89C51ED2: --code-loc 0x0000 --iram-size 256 --xram-loc 0x0000 --xram-size 0x0700
// AT89S8253:  --code-loc 0x0000 --iram-size 256
// ********************************************************************
// Fr 8051-Microcontroller, z.B. AT89C51ED2 oder AT89S8253
// Frequenz: 11.0592 MHz
// Code-Speicherbedarf: ca. 0.8kByte, keine XRAM-Verwendung
// Das Programm verwendet nur 8051/52 Standard-SFRs und wurde
// mit dem AT89C51ED2 getestet.
// ********************************************************************
// Die Programmfunktionen drfen fr private Nutzung verwendet und
// verndert werden. Fr gewerbliche Nutzung ist die Zustimmung
// der Authoren erforderlich. Die Nutzung erfolgt auf eigene Gefahr.
// Fr eventuell entstehende Schden wird keine Haftung bernommen.
// ********************************************************************

#include "at89c51ed2.h"   // Register des AT89C51ED2
//#include "at89s8253.h"   // Register des AT89S8253
#include <string.h>      // Stingfunktionen

// Typendefinition
typedef unsigned char byte;
typedef unsigned int  word;
typedef unsigned long dword;

// Definebereich
#define LCD_ON      P1_4
#define LCD_E1      P3_5
#define LCD_RS      P3_7
#define LCD_RW      P3_6
#define LCD_DATA    P2

#define SPALTEN_MAX  16
#define ZEILEN_MAX   2

// ********************************************************************
// Verzgerungsschleife fr kurze Zeit
// Zeit: Wartezeit in [zeit]
//       1 = 20us
//       2 = 28us
//       3 = 36us u.s.w. (gilt fr 11.0952MHz)
//     255 = ca. 2ms
// ********************************************************************
void delay_us(byte zeit) {
byte idata zaehler;
  for ( zaehler = zeit; zaehler; zaehler-- ) {
   _asm nop; _endasm;
  }
}

// ********************************************************************
// Verzgerungsschleife fr lange Zeit
// Zeit: Wartezeit in [zeit]
//       1 = ca. 10ms
//       2 = ca. 20ms
//       3 = ca. 30ms u.s.w.
//     255 = ca. 2,5s (gilt fr 11.0952MHz)
// ********************************************************************
void delay(byte zeit) {
byte idata zaehler;
  for (zaehler = zeit; zaehler; zaehler--) {
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
  }
}

// ***********************************************************************
// Schreiben eines Zeichens an das LCD-Modul
// bergabe: lcd_byte : Auszugebendes Zeichen/Steuerzeichen
//           lcd_mode : 0 - Daten
//                      1 - Steuerzeichen
// ***********************************************************************
void write_lcd(byte lcd_byte, bit lcd_mode) {
byte idata lcd_busy;
  // Pause lassen, damit LCD intern den letzten Befehl abarbeiten kann
  for(lcd_busy = 50; lcd_busy; lcd_busy--); // Pause mind. 26,3us
  // Hier knnte man alternativ testen, ob LCD-Modul noch im internen
  // Zyklus ist (Busy-Flag abfragen). Beachte! Durch die COG-Technik
  // kann nur ein sehr geringer Ausgangsstrom geliefert werden, sodass
  // nicht alle Portpins sicher auf H gezogen werden. Ggf. muss
  // ein Treiber oder ein sehr hochohmiger Eingang verwendet werden.

  // LCD-Anzeige ist zum Empfang neuer Zeichen bereit
  LCD_DATA = lcd_byte; // Byte ausgeben
  LCD_RW = 0;
  if (lcd_mode) LCD_RS = 0; // Steuerzeichen
    else LCD_RS = 1; // Daten
  LCD_E1 = 1;
  _asm nop; nop; nop; _endasm;     // Impulsverzgerung
  LCD_E1 = 0;
}

// ***********************************************************************
// Lschen des LCD-Display
//         Kursor auf Position 0,0
// ***********************************************************************
void clear_lcd(void) {
  write_lcd(0x01,1);
  delay_us(200);     // Wartezeit, weil Clear-Befehl lnger dauert
} 

// ***********************************************************************
// Ausgabe eines ASCII-Zeichen positioniert auf dem LCD-Modul
// bergabe: lcd_x : Spalte (0...SPALTEN_MAX-1)
//           lcd_y : Zeile  (0...ZEILEN_MAX-1)
//           lcd_ascii : ASCII-Zeichen
// *********************************************************************** 
void write_byte_lcd(byte lcd_x, byte lcd_y, byte lcd_ascii) {
byte idata lcd_offset = 0;
  if (lcd_x > (SPALTEN_MAX - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN_MAX  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x80; break;  // Zeile 1
    case 1:  lcd_offset = 0xC0; break;  // Zeile 2 
    default: lcd_offset = 0x80;         // Zeile>2
  };
  write_lcd(lcd_x+lcd_offset,1);        // Kursorposition setzen
  write_lcd(lcd_ascii,0);               // Ausgabe des ASCII-Zeichen an der Kursorposition
}

// ***********************************************************************
// Ausgabe einer Zeichenkette positioniert auf dem LCD-Modul
// bergabe: lcd_x : Spalte (0...SPALTEN_MAX-1)
//           lcd_y : Zeile  (0...ZEILEN_MAX-1)
//           lcd_zeichen : Adresse der auszugebenden format.  Zeichenkette
//           clr_line    : Lschen bis Zeilenende
//                         1 - Lschen
//                         0 - kein Lschen
// ***********************************************************************
void printf_lcd(byte lcd_x, byte lcd_y, byte *lcd_zeichen, bit clr_line) {
byte idata lcd_i;
byte idata lcd_offset = 0;
  if (lcd_x > (SPALTEN_MAX - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN_MAX  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x80; break;  // Zeile 1
    case 1:  lcd_offset = 0xC0; break;  // Zeile 2 
    default: lcd_offset = 0x80;         // Zeile>2
  }
  write_lcd(lcd_x+lcd_offset,1);        // Kursorposition setzen
  // Ausgabe der Zeichenkette ab der Kursorposition
  lcd_offset = strlen(lcd_zeichen);     // Lnge der Zeichenkette
  if (lcd_offset > SPALTEN_MAX) lcd_offset = SPALTEN_MAX;
  for(lcd_i = lcd_offset ; lcd_i ; lcd_i--) {
    write_lcd(*lcd_zeichen,0); lcd_zeichen++;
  }
  if (clr_line) {
    // Lschen bis Zeilenende
    for(lcd_i = SPALTEN_MAX - lcd_offset - lcd_x ; lcd_i ; lcd_i--)
      write_lcd(' ',0);
  }
}

// ***********************************************************************
// Blinkposition auf dem LCD-Displays setzen / Blinken ein/ausschalten
// bergabe: lcd_x : Spalte (0...SPALTEN_MAX-1)
//           lcd_y : Zeile  (0...ZEILEN_MAX-1)
//           lcd_blink :   0 - Blinken/Cursor aus
//                         1 - Blinken an
//                         2 - Cursor  an
// ***********************************************************************
void blink_lcd(byte lcd_x, byte lcd_y, byte lcd_blink) {
byte idata lcd_offset;
  write_lcd(0x0C,1);                     // KURSOR ausschalten
  if (lcd_x > (SPALTEN_MAX - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN_MAX  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x80; break;  // Zeile 1
    case 1:  lcd_offset = 0xC0; break;  // Zeile 2 
    default: lcd_offset = 0x80;         // Zeile>2
  };
  write_lcd(lcd_x+lcd_offset,1);         // Blinkposition setzen
  if (lcd_blink == 1) write_lcd(0x0D,1); // Blinken ein
  if (lcd_blink == 2) write_lcd(0x0E,1); // Cursor ein
}


// ***********************************************************************
// Grundinitialisierung des LCD-Moduls
// ***********************************************************************
void init_lcd(void) {
byte i_n;
  delay(5); // ca. 50ms Wartezeit nach dem Einschalten 
  LCD_DATA = 0x39;     // 39 Initialisierung 2x 0011 1001
  LCD_RW = 0;
  LCD_RS = 0;
  for( i_n = 2; i_n; i_n--){ 	   
    delay(5);
    LCD_E1 = 1;
    _asm nop; nop; nop; _endasm;     // Impulsverzgerung
    LCD_E1 = 0;
  }
  // Grundinitialisierung
  write_lcd(0x1C,1);               // BS: 1/4, 2 Zeilen
  write_lcd(0x52,1);               // 52 Boster aus, Kontrast C5,C4
  write_lcd(0x69,1);               // 69 Spannungsfolger und Verstrkung
  write_lcd(0x74,1);               // Kontrast C3,C2,C1
  write_lcd(0x0F,1);               // Display ein, Cursor ein, Cursor blinken
  write_lcd(0x06,1);               // Autoincrement
  clear_lcd();
}

// ********************************************************************
// Grund-Initialisierung
// ********************************************************************
void init_system (void) {
  LCD_ON   = 1; // Beleuchtung LCD einschalten
  LCD_E1   = 0; // Enable-Pin LCD
}

// ********************************************************************
// Hauptprogramm
// ********************************************************************
void main (void) {
  init_system();
  init_lcd();
  printf_lcd(0,0,"LCD-Basic's",1); // Einschaltmeldung
  printf_lcd(0,1,"DOG-M 162 ST7036",1);  
  delay(200);  // Pause ca. 2s
  printf_lcd(0,1,"V1.0 17.10.2007",1);  
  delay(200);  // Pause ca. 2s
  printf_lcd(0,1,"8Bit parallel-M",1);
  blink_lcd(15,1,1); 
  while(1);
}
